@@ -5,6 +5,8 @@
 #
 # License: MIT
 
+from __future__ import division
+
 import warnings
 from random import Random
 import os
@@ -20,13 +22,14 @@ from PIL import ImageDraw
 from PIL import ImageFont
 
 from .query_integral_image import query_integral_image
+from .tokenization import unigrams_and_bigrams, process_tokens
 
 item1 = itemgetter(1)
 
 FONT_PATH = os.environ.get("FONT_PATH", os.path.join(os.path.dirname(__file__),
                                                      "DroidSansMono.ttf"))
-STOPWORDS = set([x.strip() for x in open(os.path.join(os.path.dirname(__file__),
-                                                      'stopwords')).read().split('\n')])
+STOPWORDS = set([x.strip() for x in open(
+    os.path.join(os.path.dirname(__file__), 'stopwords')).read().split('\n')])
 
 
 class IntegralOccupancyMap(object):
@@ -41,11 +44,12 @@ class IntegralOccupancyMap(object):
             self.integral = np.zeros((height, width), dtype=np.uint32)
 
     def sample_position(self, size_x, size_y, random_state):
-        return query_integral_image(self.integral, size_x, size_y, random_state)
+        return query_integral_image(self.integral, size_x, size_y,
+                                    random_state)
 
     def update(self, img_array, pos_x, pos_y):
-        partial_integral = np.cumsum(np.cumsum(img_array[pos_x:, pos_y:], axis=1),
-                                     axis=0)
+        partial_integral = np.cumsum(np.cumsum(img_array[pos_x:, pos_y:],
+                                               axis=1), axis=0)
         # paste recomputed part into old image
         # if x or y is zero it is a bit annoying
         if pos_x > 0:
@@ -72,7 +76,8 @@ def random_color_func(word=None, font_size=None, position=None,
     word, font_size, position, orientation  : ignored.
 
     random_state : random.Random object or None, (default=None)
-        If a random object is given, this is used for generating random numbers.
+        If a random object is given, this is used for generating random
+        numbers.
 
     """
     if random_state is None:
@@ -80,16 +85,43 @@ def random_color_func(word=None, font_size=None, position=None,
     return "hsl(%d, 80%%, 50%%)" % random_state.randint(0, 255)
 
 
+class colormap_color_func(object):
+    """Color func created from matplotlib colormap.
+
+    Parameters
+    ----------
+    colormap : string or matplotlib colormap
+        Colormap to sample from
+
+    Example
+    -------
+    >>> WordCloud(color_func=colormap_color_func("magma"))
+
+    """
+    def __init__(self, colormap):
+        import matplotlib.pyplot as plt
+        self.colormap = plt.cm.get_cmap(colormap)
+
+    def __call__(self, word, font_size, position, orientation,
+                 random_state=None, **kwargs):
+        if random_state is None:
+            random_state = Random()
+        r, g, b, _ = 255 * np.array(self.colormap(random_state.uniform(0, 1)))
+        return "rgb({:.0f}, {:.0f}, {:.0f})".format(r, g, b)
+
+
 def get_single_color_func(color):
     """Create a color function which returns a single hue and saturation with.
-    different values (HSV). Accepted values are color strings as usable by PIL/Pillow.
+    different values (HSV). Accepted values are color strings as usable by
+    PIL/Pillow.
 
     >>> color_func1 = get_single_color_func('deepskyblue')
     >>> color_func2 = get_single_color_func('#00b4d2')
     """
     old_r, old_g, old_b = ImageColor.getrgb(color)
     rgb_max = 255.
-    h, s, v = colorsys.rgb_to_hsv(old_r / rgb_max, old_g / rgb_max, old_b / rgb_max)
+    h, s, v = colorsys.rgb_to_hsv(old_r / rgb_max, old_g / rgb_max,
+                                  old_b / rgb_max)
 
     def single_color_func(word=None, font_size=None, position=None,
                           orientation=None, font_path=None, random_state=None):
@@ -103,13 +135,15 @@ def get_single_color_func(color):
         word, font_size, position, orientation  : ignored.
 
         random_state : random.Random object or None, (default=None)
-          If a random object is given, this is used for generating random numbers.
+          If a random object is given, this is used for generating random
+          numbers.
 
         """
         if random_state is None:
             random_state = Random()
         r, g, b = colorsys.hsv_to_rgb(h, s, random_state.uniform(0.2, 1))
-        return 'rgb({:.0f}, {:.0f}, {:.0f})'.format(r * rgb_max, g * rgb_max, b * rgb_max)
+        return 'rgb({:.0f}, {:.0f}, {:.0f})'.format(r * rgb_max, g * rgb_max,
+                                                    b * rgb_max)
     return single_color_func
 
 
@@ -131,6 +165,9 @@ class WordCloud(object):
 
     prefer_horizontal : float (default=0.90)
         The ratio of times to try horizontal fitting as opposed to vertical.
+        If prefer_horizontal < 1, the algorithm will try rotating the word
+        if it doesn't fit. (There is currently no built-in way to get only
+        vertical words.)
 
     mask : nd-array or None (default=None)
         If not None, gives a binary mask on where to draw words. If mask is not
@@ -155,8 +192,9 @@ class WordCloud(object):
     max_words : number (default=200)
         The maximum number of words.
 
-    stopwords : set of strings
-        The words that will be eliminated.
+    stopwords : set of strings or None
+        The words that will be eliminated. If None, the build-in STOPWORDS
+        list will be used.
 
     background_color : color value (default="black")
         Background color for the word cloud image.
@@ -169,30 +207,59 @@ class WordCloud(object):
         Transparent background will be generated when mode is "RGBA" and
         background_color is None.
 
-    relative_scaling : float (default=0)
-        Importance of relative word frequencies for font-size.
-        With relative_scaling=0, only word-ranks are considered.
-        With relative_scaling=1, a word that is twice as frequent will have twice the size.
-        If you want to consider the word frequencies and not only their rank, relative_scaling
-        around .5 often looks good.
+    relative_scaling : float (default=.5)
+        Importance of relative word frequencies for font-size.  With
+        relative_scaling=0, only word-ranks are considered.  With
+        relative_scaling=1, a word that is twice as frequent will have twice
+        the size.  If you want to consider the word frequencies and not only
+        their rank, relative_scaling around .5 often looks good.
+
+        .. versionchanged: 2.0
+            Default is now 0.5.
+
+    color_func : callable, default=None
+        Callable with parameters word, font_size, position, orientation,
+        font_path, random_state that returns a PIL color for each word.
+        Overwrites "colormap".
+        See colormap for specifying a matplotlib colormap instead.
 
     regexp : string or None (optional)
         Regular expression to split the input text into tokens in process_text.
         If None is specified, ``r"\w[\w']+"`` is used.
 
+    collocations : bool, default=True
+        Whether to include collocations (bigrams) of two words.
+
+        .. versionadded: 2.0
+
+    colormap : string or matplotlib colormap, default="viridis"
+        Matplotlib colormap to randomly draw colors from for each word.
+        Ignored if "color_func" is specified.
+
+        .. versionadded: 2.0
+
+    normalize_plurals : bool, default=True
+        Whether to remove trailing 's' from words. If True and a word
+        appears with and without a trailing 's', the one with trailing 's'
+        is removed and its counts are added to the version without
+        trailing 's' -- unless the word ends with 'ss'.
+
     Attributes
     ----------
-    ``words_``: list of tuples (string, float)
+    ``words_`` : dict of string to float
         Word tokens with associated frequency.
 
+        .. versionchanged: 2.0
+            ``words_`` is now a dictionary
+
     ``layout_`` : list of tuples (string, int, (int, int), int, color))
         Encodes the fitted word cloud. Encodes for each word the string, font
         size, position, orientation and color.
 
     Notes
     -----
-    Larger canvases with make the code significantly slower. If you need a large
-    word cloud, try a lower canvas size, and set the scale parameter.
+    Larger canvases with make the code significantly slower. If you need a
+    large word cloud, try a lower canvas size, and set the scale parameter.
 
     The algorithm might give more weight to the ranking of the words
     than their actual frequencies, depending on the ``max_font_size`` and the
@@ -200,12 +267,24 @@ class WordCloud(object):
     """
 
     def __init__(self, font_path=None, width=400, height=200, margin=2,
-                 ranks_only=None, prefer_horizontal=0.9, mask=None, scale=1,
-                 color_func=random_color_func, max_words=200, min_font_size=4,
+                 ranks_only=None, prefer_horizontal=.9, mask=None, scale=1,
+                 color_func=None, max_words=200, min_font_size=4,
                  stopwords=None, random_state=None, background_color='black',
-                 max_font_size=None, font_step=1, mode="RGB", relative_scaling=0, regexp=None):
+                 max_font_size=None, font_step=1, mode="RGB",
+                 relative_scaling=.5, regexp=None, collocations=True,
+                 colormap=None, normalize_plurals=True):
         if font_path is None:
             font_path = FONT_PATH
+        if color_func is None and colormap is None:
+            # we need a color map
+            import matplotlib
+            version = matplotlib.__version__
+            if version[0] < "2" and version[2] < "5":
+                colormap = "hsv"
+            else:
+                colormap = "viridis"
+        self.colormap = colormap
+        self.collocations = collocations
         self.font_path = font_path
         self.width = width
         self.height = height
@@ -213,9 +292,9 @@ class WordCloud(object):
         self.prefer_horizontal = prefer_horizontal
         self.mask = mask
         self.scale = scale
-        self.color_func = color_func
+        self.color_func = color_func or colormap_color_func(colormap)
         self.max_words = max_words
-        self.stopwords = stopwords or STOPWORDS
+        self.stopwords = stopwords if stopwords is not None else STOPWORDS
         self.min_font_size = min_font_size
         self.font_step = font_step
         self.regexp = regexp
@@ -223,17 +302,17 @@ class WordCloud(object):
             random_state = Random(random_state)
         self.random_state = random_state
         self.background_color = background_color
-        if max_font_size is None:
-            max_font_size = height
         self.max_font_size = max_font_size
         self.mode = mode
         if relative_scaling < 0 or relative_scaling > 1:
-            raise ValueError("relative_scaling needs to be between 0 and 1, got %f."
-                             % relative_scaling)
+            raise ValueError("relative_scaling needs to be "
+                             "between 0 and 1, got %f." % relative_scaling)
         self.relative_scaling = relative_scaling
         if ranks_only is not None:
             warnings.warn("ranks_only is deprecated and will be removed as"
-                          " it had no effect. Look into relative_scaling.", DeprecationWarning)
+                          " it had no effect. Look into relative_scaling.",
+                          DeprecationWarning)
+        self.normalize_plurals = normalize_plurals
 
     def fit_words(self, frequencies):
         """Create a word_cloud from words and frequencies.
@@ -242,8 +321,8 @@ class WordCloud(object):
 
         Parameters
         ----------
-        frequencies : array of tuples
-            A tuple contains the word and its frequency.
+        frequencies : dict from string to float
+            A contains words and associated frequency.
 
         Returns
         -------
@@ -251,13 +330,16 @@ class WordCloud(object):
         """
         return self.generate_from_frequencies(frequencies)
 
-    def generate_from_frequencies(self, frequencies):
+    def generate_from_frequencies(self, frequencies, max_font_size=None):
         """Create a word_cloud from words and frequencies.
 
         Parameters
         ----------
-        frequencies : array of tuples
-            A tuple contains the word and its frequency.
+        frequencies : dict from string to float
+            A contains words and associated frequency.
+
+        max_font_size : int
+            Use this font-size instead of self.max_font_size
 
         Returns
         -------
@@ -265,38 +347,38 @@ class WordCloud(object):
 
         """
         # make sure frequencies are sorted and normalized
-        frequencies = sorted(frequencies, key=item1, reverse=True)
+        frequencies = sorted(frequencies.items(), key=item1, reverse=True)
+        if len(frequencies) <= 0:
+            raise ValueError("We need at least 1 word to plot a word cloud, "
+                             "got %d." % len(frequencies))
         frequencies = frequencies[:self.max_words]
+
         # largest entry will be 1
         max_frequency = float(frequencies[0][1])
 
-        frequencies = [(word, freq / max_frequency) for word, freq in frequencies]
-
-        self.words_ = frequencies
+        frequencies = [(word, freq / max_frequency)
+                       for word, freq in frequencies]
 
         if self.random_state is not None:
             random_state = self.random_state
         else:
             random_state = Random()
 
-        if len(frequencies) <= 0:
-            print("We need at least 1 word to plot a word cloud, got %d."
-                  % len(frequencies))
-
         if self.mask is not None:
             mask = self.mask
             width = mask.shape[1]
             height = mask.shape[0]
             if mask.dtype.kind == 'f':
-                warnings.warn("mask image should be unsigned byte between 0 and"
-                              " 255. Got a float array")
+                warnings.warn("mask image should be unsigned byte between 0"
+                              " and 255. Got a float array")
             if mask.ndim == 2:
                 boolean_mask = mask == 255
             elif mask.ndim == 3:
                 # if all channels are white, mask out
                 boolean_mask = np.all(mask[:, :, :3] == 255, axis=-1)
             else:
-                raise ValueError("Got mask of invalid shape: %s" % str(mask.shape))
+                raise ValueError("Got mask of invalid shape: %s"
+                                 % str(mask.shape))
         else:
             boolean_mask = None
             height, width = self.height, self.width
@@ -308,35 +390,68 @@ class WordCloud(object):
         img_array = np.asarray(img_grey)
         font_sizes, positions, orientations, colors = [], [], [], []
 
-        font_size = self.max_font_size
         last_freq = 1.
 
+        if max_font_size is None:
+            # if not provided use default font_size
+            max_font_size = self.max_font_size
+
+        if max_font_size is None:
+            # figure out a good font size by trying to draw with
+            # just the first two words
+            if len(frequencies) == 1:
+                # we only have one word. We make it big!
+                font_size = self.height
+            else:
+                self.generate_from_frequencies(dict(frequencies[:2]),
+                                               max_font_size=self.height)
+                # find font sizes
+                sizes = [x[1] for x in self.layout_]
+                font_size = int(2 * sizes[0] * sizes[1] / (sizes[0]
+                                                           + sizes[1]))
+        else:
+            font_size = max_font_size
+
+        # we set self.words_ here because we called generate_from_frequencies
+        # above... hurray for good design?
+        self.words_ = dict(frequencies)
+
         # start drawing grey image
         for word, freq in frequencies:
             # select the font size
             rs = self.relative_scaling
             if rs != 0:
-                font_size = int(round((rs * (freq / float(last_freq)) + (1 - rs)) * font_size))
+                font_size = int(round((rs * (freq / float(last_freq))
+                                       + (1 - rs)) * font_size))
+            if random_state.random() < self.prefer_horizontal:
+                orientation = None
+            else:
+                orientation = Image.ROTATE_90
+            tried_other_orientation = False
             while True:
                 # try to find a position
                 font = ImageFont.truetype(self.font_path, font_size)
                 # transpose font optionally
-                if random_state.random() < self.prefer_horizontal:
-                    orientation = None
-                else:
-                    orientation = Image.ROTATE_90
-                transposed_font = ImageFont.TransposedFont(font,
-                                                           orientation=orientation)
+                transposed_font = ImageFont.TransposedFont(
+                    font, orientation=orientation)
                 # get size of resulting text
                 box_size = draw.textsize(word, font=transposed_font)
                 # find possible places using integral image:
                 result = occupancy.sample_position(box_size[1] + self.margin,
                                                    box_size[0] + self.margin,
                                                    random_state)
-                if result is not None or font_size == 0:
+                if result is not None or font_size < self.min_font_size:
+                    # either we found a place or font-size went too small
                     break
                 # if we didn't find a place, make font smaller
-                font_size -= self.font_step
+                # but first try to rotate!
+                if not tried_other_orientation and self.prefer_horizontal < 1:
+                    orientation = (Image.ROTATE_90 if orientation is None else
+                                   Image.ROTATE_90)
+                    tried_other_orientation = True
+                else:
+                    font_size -= self.font_step
+                    orientation = None
 
             if font_size < self.min_font_size:
                 # we were unable to draw any more
@@ -363,7 +478,8 @@ class WordCloud(object):
             occupancy.update(img_array, x, y)
             last_freq = freq
 
-        self.layout_ = list(zip(frequencies, font_sizes, positions, orientations, colors))
+        self.layout_ = list(zip(frequencies, font_sizes, positions,
+                                orientations, colors))
         return self
 
     def process_text(self, text):
@@ -388,53 +504,35 @@ class WordCloud(object):
         include all those things.
         """
 
-        self.stopwords_lower_ = set(map(str.lower, self.stopwords))
+        stopwords = set([i.lower() for i in self.stopwords])
 
-        d = {}
         flags = (re.UNICODE if sys.version < '3' and type(text) is unicode
                  else 0)
         regexp = self.regexp if self.regexp is not None else r"\w[\w']+"
-        for word in re.findall(regexp, text, flags=flags):
-            if word.isdigit():
-                continue
 
-            word_lower = word.lower()
-            if word_lower in self.stopwords_lower_:
-                continue
+        words = re.findall(regexp, text, flags)
+        # remove stopwords
+        words = [word for word in words if word.lower() not in stopwords]
+        # remove 's
+        words = [word[:-2] if word.lower().endswith("'s") else word
+                 for word in words]
+        # remove numbers
+        words = [word for word in words if not word.isdigit()]
 
-            # Look in lowercase dict.
-            try:
-                d2 = d[word_lower]
-            except KeyError:
-                d2 = {}
-                d[word_lower] = d2
+        if self.collocations:
+            word_counts = unigrams_and_bigrams(words, self.normalize_plurals)
+        else:
+            word_counts, _ = process_tokens(words, self.normalize_plurals)
 
-            # Look in any case dict.
-            d2[word] = d2.get(word, 0) + 1
-
-        # merge plurals into the singular count (simple cases only)
-        for key in list(d.keys()):
-            if key.endswith('s'):
-                key_singular = key[:-1]
-                if key_singular in d:
-                    dict_plural = d[key]
-                    dict_singular = d[key_singular]
-                    for word, count in dict_plural.items():
-                        singular = word[:-1]
-                        dict_singular[singular] = dict_singular.get(singular, 0) + count
-                    del d[key]
-
-        d3 = {}
-        for d2 in d.values():
-            # Get the most popular case.
-            first = max(d2.items(), key=item1)[0]
-            d3[first] = sum(d2.values())
-
-        return d3
+        return word_counts
 
     def generate_from_text(self, text):
         """Generate wordcloud from text.
 
+        The input "text" is expected to be a natural text. If you pass a sorted
+        list of words, words will appear in your output twice. To remove this
+        duplication, set ``collocations=False``.
+
         Calls process_text and generate_from_frequencies.
 
         ..versionchanged:: 1.2.2
@@ -446,12 +544,16 @@ class WordCloud(object):
         self
         """
         words = self.process_text(text)
-        self.generate_from_frequencies(words.items())
+        self.generate_from_frequencies(words)
         return self
 
     def generate(self, text):
         """Generate wordcloud from text.
 
+        The input "text" is expected to be a natural text. If you pass a sorted
+        list of words, words will appear in your output twice. To remove this
+        duplication, set ``collocations=False``.
+
         Alias to generate_from_text.
 
         Calls process_text and generate_from_frequencies.
@@ -465,7 +567,8 @@ class WordCloud(object):
     def _check_generated(self):
         """Check if ``layout_`` was computed, otherwise raise error."""
         if not hasattr(self, "layout_"):
-            raise ValueError("WordCloud has not been calculated, call generate first.")
+            raise ValueError("WordCloud has not been calculated, call generate"
+                             " first.")
 
     def to_image(self):
         self._check_generated()
@@ -475,21 +578,25 @@ class WordCloud(object):
         else:
             height, width = self.height, self.width
 
-        img = Image.new(self.mode, (int(width * self.scale), int(height * self.scale)),
+        img = Image.new(self.mode, (int(width * self.scale),
+                                    int(height * self.scale)),
                         self.background_color)
         draw = ImageDraw.Draw(img)
         for (word, count), font_size, position, orientation, color in self.layout_:
-            font = ImageFont.truetype(self.font_path, int(font_size * self.scale))
-            transposed_font = ImageFont.TransposedFont(font,
-                                                       orientation=orientation)
-            pos = (int(position[1] * self.scale), int(position[0] * self.scale))
+            font = ImageFont.truetype(self.font_path,
+                                      int(font_size * self.scale))
+            transposed_font = ImageFont.TransposedFont(
+                font, orientation=orientation)
+            pos = (int(position[1] * self.scale),
+                   int(position[0] * self.scale))
             draw.text(pos, word, fill=color, font=transposed_font)
         return img
 
-    def recolor(self, random_state=None, color_func=None):
+    def recolor(self, random_state=None, color_func=None, colormap=None):
         """Recolor existing layout.
 
-        Applying a new coloring is much faster than generating the whole wordcloud.
+        Applying a new coloring is much faster than generating the whole
+        wordcloud.
 
         Parameters
         ----------
@@ -501,6 +608,10 @@ class WordCloud(object):
             Function to generate new color from word count, font size, position
             and orientation.  If None, self.color_func is used.
 
+        colormap : string or matplotlib colormap, default=None
+            Use this colormap to generate new colors. Ignored if color_func
+            is specified. If None, self.color_func (or self.color_map) is used.
+
         Returns
         -------
         self
@@ -510,12 +621,17 @@ class WordCloud(object):
         self._check_generated()
 
         if color_func is None:
-            color_func = self.color_func
+            if colormap is None:
+                color_func = self.color_func
+            else:
+                color_func = colormap_color_func(colormap)
         self.layout_ = [(word_freq, font_size, position, orientation,
                          color_func(word=word_freq[0], font_size=font_size,
                                     position=position, orientation=orientation,
-                                    random_state=random_state, font_path=self.font_path))
-                        for word_freq, font_size, position, orientation, _ in self.layout_]
+                                    random_state=random_state,
+                                    font_path=self.font_path))
+                        for word_freq, font_size, position, orientation, _
+                        in self.layout_]
         return self
 
     def to_file(self, filename):
